package example.localplugin.coolemall.computebox1;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import schedframe.events.scheduling.SchedulingEvent;
import schedframe.exceptions.ResourceException;
import schedframe.resources.ResourceStatus;
import schedframe.resources.StandardResourceType;
import schedframe.resources.computing.Node;
import schedframe.resources.computing.ComputingResource;
import schedframe.resources.computing.Core;
import schedframe.resources.computing.Processor;
import schedframe.resources.computing.profiles.energy.airthroughput.UserAirThroughputStateName;
import schedframe.resources.computing.profiles.energy.power.StandardPowerStateName;
import schedframe.resources.devices.Device;
import schedframe.resources.units.ProcessingElements;
import schedframe.resources.units.ResourceUnit;
import schedframe.resources.units.ResourceUnitName;
import schedframe.resources.units.StandardResourceUnitName;
import schedframe.scheduling.manager.resources.ClusterResourceManager;
import schedframe.scheduling.manager.resources.ResourceManager;
import schedframe.scheduling.manager.tasks.JobRegistry;
import schedframe.scheduling.manager.tasks.JobRegistryImpl;
import schedframe.scheduling.plan.SchedulingPlanInterface;
import schedframe.scheduling.plan.impl.SchedulingPlan;
import schedframe.scheduling.plugin.grid.ModuleList;
import schedframe.scheduling.queue.TaskQueue;
import schedframe.scheduling.queue.TaskQueueList;
import schedframe.scheduling.tasks.TaskInterface;
import test.Node_Fan_Mapping;
import example.localplugin.BaseLocalSchedulingPlugin;
import gridsim.dcworms.DCWormsTags;

public class CB1_FCFS_ConsolidationAtom_NodePowMan_SP extends BaseLocalSchedulingPlugin {

	
	public CB1_FCFS_ConsolidationAtom_NodePowMan_SP () {
		Node_Fan_Mapping.init();
	}

	public SchedulingPlanInterface<?> schedule(SchedulingEvent event, TaskQueueList queues, JobRegistry jobRegistry,
			 ResourceManager resManager, ModuleList modules) {

		ClusterResourceManager resourceManager = (ClusterResourceManager) resManager;
		SchedulingPlan plan = new SchedulingPlan();
		// choose the events types to serve.
		// Different actions for different events are possible.
		switch (event.getType()) {
		case START_TASK_EXECUTION:
		case TASK_FINISHED:
		//case TIMER:
			// our tasks are placed only in first queue (see
			// BaseLocalSchedulingPlugin.placeJobsInQueues() method)
			TaskQueue q = queues.get(0);
			
			List<Node> notSelectedNodes = null;
			try {
				notSelectedNodes = (List<Node>) resourceManager.getResourcesOfType(StandardResourceType.Node);
			} catch (ResourceException e) {
			}
			
			// check all tasks in queue
			for (int i = 0; i < q.size(); i++) {
				TaskInterface<?> task = q.get(i);
				// if status of the tasks in READY
				if (task.getStatus() == DCWormsTags.READY) {

					Map<ResourceUnitName, ResourceUnit> choosenResources = chooseResourcesForExecution(resourceManager, task);
					if (choosenResources != null) {
						addToSchedulingPlan(plan, task, choosenResources);
						ProcessingElements pe = (ProcessingElements) choosenResources.get(StandardResourceUnitName.PE);
						Node node = (Node) pe.get(0).getParent().getParent();
						notSelectedNodes.remove(node);
					}
				}
			}
			manageResources(notSelectedNodes);
			break;
		}
		return plan;
	}

	@SuppressWarnings("unchecked")
	private Map<ResourceUnitName, ResourceUnit> chooseResourcesForExecution(
			ClusterResourceManager resourceManager, TaskInterface<?> task) {

		Map<ResourceUnitName, ResourceUnit> map = new HashMap<ResourceUnitName, ResourceUnit>();
		
		List<Node> nodes = null;
		try {
			nodes = (List<Node>) resourceManager.getResourcesOfType(StandardResourceType.Node);
		} catch (ResourceException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		List<Node> avNodes = filterNodes(nodes, task);
		Node node;
		if(avNodes.size() == 0){
			try {
				nodes = (List<Node>) resourceManager.getResourcesByTypeWithStatus(StandardResourceType.Node, ResourceStatus.UNAVAILABLE);
			} catch (ResourceException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			Collections.sort(nodes, new FavourableNodeComparator());
			node = turnOnFirstNode(nodes, task);
			if(node == null)
				return null;
		} else {
			Collections.sort(avNodes, new FavourableNodeComparator());
			node = findTheMostLoadedNode(avNodes);			
		}

		int cpuRequest;
		try {
			cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
		} catch (NoSuchFieldException e) {
			cpuRequest = 0;
		}

		if (cpuRequest != 0) {

			List<ComputingResource> choosenResources = new ArrayList<ComputingResource>();	
			
			List<Core> cores = node.getCores();			
			for (int i = 0; i < cores.size() && cpuRequest > 0; i++) {
				if (cores.get(i).getStatus() == ResourceStatus.FREE) {
					choosenResources.add(cores.get(i));
					cpuRequest--;
				}
			}
			if (cpuRequest > 0) {
				return null;
			}
			//choosenResources.add(node.getProcessors().get(0));
			ProcessingElements pe = new ProcessingElements();
			pe.addAll(choosenResources);
			map.put(StandardResourceUnitName.PE, pe);
			return map;
		}
		return null;
	}
	
	private List<Node> filterNodes(List<Node> nodes, TaskInterface<?> task){
		List<Node> filteredNodes = new ArrayList<Node>();
		int cpuRequest;
		try {
			cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
		} catch (NoSuchFieldException e) {
			cpuRequest = 0;
		}
		for (Node node : nodes) {

			if (cpuRequest != 0) {

				List<Core> cores = node.getCores();
				if (cores.size() < cpuRequest) {
					if(cores.size() == 0){
						if(node.getProcessors().size() < cpuRequest)
							continue;
					}
				}

				int freeCores = 0;
				for(Core core: cores){
					if(core.getStatus() == ResourceStatus.FREE)
						freeCores++;
				}
				
				if(freeCores < cpuRequest)
					continue;
				
				filteredNodes.add(node);
			}
		}
		
		return filteredNodes;
	}
	
	private Node findTheMostLoadedNode(List<Node> nodes){
		Node mostLoadedNode = null;
			
		int maxRunningTask = -1;
		int nrRunningTasks;
		for(Node node: nodes){
			JobRegistry jr = new JobRegistryImpl(node.getFullName());
			nrRunningTasks = jr.getRunningTasks().size(); 
			if(nrRunningTasks > maxRunningTask){
				mostLoadedNode = node;
				maxRunningTask = nrRunningTasks;
			}
		}
		return mostLoadedNode;
	}
	
	private Node turnOnFirstNode(List<Node> nodes, TaskInterface<?> task){
		Node startedNode = null;

		int cpuRequest;
		try {
			cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
		} catch (NoSuchFieldException e) {
			cpuRequest = 0;
		}
		
		for(Node node: nodes){
			
			if (cpuRequest != 0) {

				List<Core> cores = node.getCores();
				if (cores.size() < cpuRequest) {
					if(cores.size() == 0){
						if(node.getProcessors().size() < cpuRequest)
							continue;
					}
				}
				
				int freeCores = 0;
				for(Core core: cores){
					if(core.getStatus() == ResourceStatus.FREE || core.getStatus() == ResourceStatus.UNAVAILABLE)
						freeCores++;
				}
				
				if(freeCores < cpuRequest)
					continue;
				else {
					for(Device device: node.getParent().getResourceCharacteristic().getDevices()){
						if(device.getFullName().equals(Node_Fan_Mapping.getNode_fan().get(node.getFullName()))){
							device.getAirThroughputInterface().setAirThroughputState(new UserAirThroughputStateName("2"));
							break;
						}
					}
					node.getPowerInterface().setPowerState(StandardPowerStateName.ON);
					startedNode = node;
					break;
				}				
			}
		}
		return startedNode;
	}
	
	protected void manageResources(List<Node> notSelectedNodes) {
		turnOffIdleNodes(notSelectedNodes);
		//adjustFans(notSelectedNodes);
	
	}

	private void turnOffIdleNodes(List<Node> nodes){
		for(Node node : nodes){
			Processor proc = node.getProcessors().get(0);
			int freeCores = 0;
			for(Core core: proc.getCores()){
				if(core.getStatus() == ResourceStatus.FREE)
					freeCores++;
			}
			
			if(freeCores == proc.getCores().size()) {
				for(Device device: node.getParent().getResourceCharacteristic().getDevices()){
					if(device.getFullName().equals(Node_Fan_Mapping.getNode_fan().get(node.getFullName()))){
						device.getAirThroughputInterface().setAirThroughputState(new UserAirThroughputStateName("1"));
						break;
					}
				}
				node.getPowerInterface().setPowerState(StandardPowerStateName.OFF);
			}

		}
	}

	/*private void adjustFans(List<Node> nodes) {
		for (Node node : nodes) {

			if (node.getPowerInterface().getPowerState().equals(StandardPowerStateName.OFF)) {
				for(Device device: node.getParent().getResourceCharacteristic().getDevices()){
					if(device.getFullName().equals(Node_Fan_Mapping.getNode_fan().get(node.getFullName()))){
						device.getAirThroughputInterface().setAirThroughputState(new UserAirThroughputStateName("1"));
						break;
					}
				}
			} else {
				for(Device device: node.getParent().getResourceCharacteristic().getDevices()){
					if(device.getFullName().equals(Node_Fan_Mapping.getNode_fan().get(node.getFullName()))){
						device.getAirThroughputInterface().setAirThroughputState(new UserAirThroughputStateName("2"));
						break;
					}
				}
			}
		}
	}
	*/
	
	class FavourableNodeComparator implements Comparator<Node>{
		
	    public int compare(Node node1, Node node2){    
	    	double node1Rank = Double.MAX_VALUE;
	    	double node2Rank = Double.MAX_VALUE;
	    	
	    	try {
				String productName = node1.getProcessors().get(0).getResourceCharacteristic().getParameters().get("product").get(0).getContent();
				if(productName.equals("Fusion G - T40N"))
					node1Rank = 3;
				else if(productName.equals("Atom - D510"))
					node1Rank = 2;
				else if(productName.equals("Atom - N2600"))
					node1Rank = 1;
				else if(productName.equals("Core i7 - 2715QE"))
					node1Rank = 4;
				else if(productName.equals("Core i7 - 3615QE"))
					node1Rank = 5;
			} catch (Exception e){
				
			}
	    	try {
				String productName = node2.getProcessors().get(0).getResourceCharacteristic().getParameters().get("product").get(0).getContent();
				if(productName.equals("Fusion G - T40N"))
					node2Rank = 3;
				else if(productName.equals("Atom - D510"))
					node2Rank = 2;
				else if(productName.equals("Atom - N2600"))
					node2Rank = 1;
				else if(productName.equals("Core i7 - 2715QE"))
					node2Rank = 4;
				else if(productName.equals("Core i7 - 3615QE"))
					node2Rank = 5;
			} catch (Exception e){
				
			}
	        if(node1Rank > node2Rank)
	        	return 1;
	        else if (node1Rank < node2Rank)
	        	return -1;
	        else return 0;	   
	    }
	}
	
}
